import os
from typing import Any

import dash_bootstrap_components as dbc
import pandas as pd
from dash import Input, Output, callback, dcc, html
from plotly.graph_objs import Figure

from ampere.app_shared import cache, update_tooltip
from ampere.common import timeit
from ampere.styling import ScreenWidth
from ampere.viz import (
    get_summary_data,
    read_plotly_fig_pickle,
    viz_summary,
)


@callback(
    Output("summary-date-slider", "tooltip"),
    Input("breakpoints", "widthBreakpoint"),
)
def update_summary_slider(breakpoint_name: str) -> dict[Any, Any]:
    return update_tooltip(breakpoint_name)


@callback(
    [
        Output("summary-date-slider", "min"),
        Output("summary-date-slider", "max"),
        Output("summary-date-slider", "value"),
        Output("summary-date-slider", "marks"),
        Output("summary-date-bounds", "data"),
    ],
    [
        Input("summary-df", "data"),
    ],
)
def get_summary_date_ranges(
    df_data: list[dict],
) -> tuple[int, int, list[int], dict[Any, dict[str, Any]], list[int]]:
    df = pd.DataFrame(df_data)
    df["metric_date"] = pd.to_datetime(df["metric_date"], utc=True)
    min_date_dt = df["metric_date"].min()
    min_date_seconds = min_date_dt.timestamp()

    max_date_dt = df["metric_date"].max()
    max_date_seconds = max_date_dt.timestamp()

    date_slider_marks = {
        min_date_seconds: {
            "label": min_date_dt.strftime("%Y-%m-%d"),
            "style": {"fontSize": 0},
        },
        max_date_seconds: {
            "label": max_date_dt.strftime("%Y-%m-%d"),
            "style": {"fontSize": 0},
        },
    }

    return (
        min_date_seconds,
        max_date_seconds,
        [min_date_seconds, max_date_seconds],
        date_slider_marks,
        [min_date_seconds, max_date_seconds],
    )


@cache.memoize()
@timeit
def get_summary_records() -> list[dict[Any, Any]]:
    return get_summary_data().to_dict("records")


@timeit
def get_viz_summary(
    df_data: list[dict],
    breakpoint_name: str,
    date_range: list[int],
    dark_mode: bool,
    date_bounds: list[int],
    metric_type: str,
) -> tuple[Figure, dict]:
    env = os.getenv("AMPERE_ENV")
    if date_range == date_bounds and env == "prod":
        mode = "dark" if dark_mode else "light"
        f_name = f"summary_{metric_type}_{mode}_{breakpoint_name}"
        try:
            fig = read_plotly_fig_pickle(f_name)
            print(f"obtained {metric_type} fig from cache")
            return fig, {}

        except Exception as e:
            print(e)
            pass

    fig = viz_summary(
        df=pd.DataFrame(df_data),
        metric_type=metric_type,
        date_range=date_range,
        screen_width=ScreenWidth(breakpoint_name),
        dark_mode=dark_mode,
    )

    return fig, {}


@callback(
    [
        Output("summary-stars", "figure"),
        Output("summary-stars", "style"),
    ],
    [
        Input("summary-df", "data"),
        Input("breakpoints", "widthBreakpoint"),
        Input("summary-date-slider", "value"),
        Input("color-mode-switch", "value"),
        Input("summary-date-bounds", "data"),
    ],
)
@timeit
def viz_summary_stars(
    df_data: list[dict],
    breakpoint_name: str,
    date_range: list[int],
    dark_mode: bool,
    date_bounds: list[int],
):
    return get_viz_summary(
        df_data, breakpoint_name, date_range, dark_mode, date_bounds, "stars"
    )


@callback(
    [
        Output("summary-issues", "figure"),
        Output("summary-issues", "style"),
    ],
    [
        Input("summary-df", "data"),
        Input("breakpoints", "widthBreakpoint"),
        Input("summary-date-slider", "value"),
        Input("color-mode-switch", "value"),
        Input("summary-date-bounds", "data"),
    ],
)
def viz_summary_issues(
    df_data: list[dict],
    breakpoint_name: str,
    date_range: list[int],
    dark_mode: bool,
    date_bounds: list[int],
):
    return get_viz_summary(
        df_data, breakpoint_name, date_range, dark_mode, date_bounds, "issues"
    )


@callback(
    [
        Output("summary-commits", "figure"),
        Output("summary-commits", "style"),
    ],
    [
        Input("summary-df", "data"),
        Input("breakpoints", "widthBreakpoint"),
        Input("summary-date-slider", "value"),
        Input("color-mode-switch", "value"),
        Input("summary-date-bounds", "data"),
    ],
)
def viz_summary_commits(
    df_data: list[dict],
    breakpoint_name: str,
    date_range: list[int],
    dark_mode: bool,
    date_bounds: list[int],
):
    return get_viz_summary(
        df_data, breakpoint_name, date_range, dark_mode, date_bounds, "commits"
    )


@callback(
    Output("summary-graph-fade", "is_in"),
    [
        Input("summary-stars", "figure"),
        Input("summary-issues", "figure"),
        Input("summary-commits", "figure"),
    ],
)
def update_summary_graph_fade(fig1, fig2, fig3):
    return all([fig1, fig2, fig3])


@callback(
    [
        Output("date-filter-width", "width"),
        Output("filter-padding-width", "width"),
        Output("filter-row", "style"),
    ],
    Input("breakpoints", "widthBreakpoint"),
)
def update_filter_for_mobile(breakpoint_name: str):
    filter_style = {"top": "60px"}
    if breakpoint_name in [ScreenWidth.xs, ScreenWidth.sm]:
        filter_style.update({"paddingTop": "20px"})
        return 11, 1, filter_style

    return 3, 8, filter_style


def layout():
    return [
        dcc.Store("summary-df", data=get_summary_records()),
        dcc.Store("summary-date-bounds"),
        dbc.Fade(
            id="summary-graph-fade",
            children=[
                dbc.Row(
                    [
                        dbc.Col(
                            html.Div(
                                dcc.RangeSlider(
                                    id="summary-date-slider",
                                    step=86400,  # daily
                                    allowCross=False,
                                ),
                                style={
                                    "whiteSpace": "nowrap",
                                    "paddingLeft": "5%",
                                    "marginTop": "6px",
                                },
                            ),
                            width=3,
                            id="date-filter-width",
                            style={"marginLeft": "5%"},
                        ),
                        dbc.Col(width=8, id="filter-padding-width"),
                    ],
                    style={"z-index": "100", "top": "60px"},
                    id="filter-row",
                ),
                html.Br(),
                dcc.Graph(
                    "summary-stars",
                    style={"visibility": "hidden"},
                    config={"displayModeBar": False},
                ),
                dcc.Graph(
                    "summary-issues",
                    style={"visibility": "hidden"},
                    config={"displayModeBar": False},
                ),
                dcc.Graph(
                    "summary-commits",
                    style={"visibility": "hidden"},
                    config={"displayModeBar": False},
                ),
            ],
            style={"transition": "opacity 200ms ease-in", "minHeight": "100vh"},
            is_in=False,
        ),
    ]



from dagster import (
    DefaultSensorStatus,
    RunFailureSensorContext,
    make_email_on_run_failure_sensor,
)
from dagster._core.execution.plan.objects import StepFailureData
from dagster._core.execution.stats import StepEventStatus

from ampere.common import get_current_time, get_secret


def generate_run_url(context: "RunFailureSensorContext") -> str:
    run_id = context.dagster_run.run_id
    run_url = f"{get_secret('AMPERE_BACKEND')}/runs/{run_id}"
    run_id_short = run_id.split("-")[0]

    return f'<h2> <a href="{run_url}">{run_id_short}</a> </h2>'


def generate_body_text(context: "RunFailureSensorContext") -> str:
    passes = 0
    pass_times = []

    run_url_str = generate_run_url(context)

    run_id = context.dagster_run.run_id
    run_step_stats = context.instance.get_run_step_stats(run_id)

    for step_stats in run_step_stats:
        if step_stats.status == StepEventStatus.SUCCESS:
            pass_times.extend([step_stats.start_time, step_stats.end_time])
            passes += 1

    pass_elapsed_time = max(pass_times) - min(pass_times)
    pass_line = f"{passes} steps passed in {pass_elapsed_time:.02f}s"

    fail_lines = []
    fails = context.get_step_failure_events()
    for fail in fails:
        event_data = fail.event_specific_data
        if not isinstance(event_data, StepFailureData):
            raise TypeError()

        error_message = str(event_data.error.cause).split("Stack Trace")[0]  # type: ignore
        fail_lines.append(
            "<br>".join([f"<h3> {fail.node_name} </h3>", error_message, "<hr>"])
        )

    output = "<br>".join(
        [
            run_url_str,
            "<hr> <h2> Passes </h2>",
            pass_line,
            "<hr> <h2> Fails </h2>",
            *fail_lines,
        ]
    )
    return output.replace("<br><br>", "<br>")


def create_email_alert_body(context: "RunFailureSensorContext") -> str:
    try:
        body_text = generate_body_text(context)
    except Exception as e:
        body_text = (
            f"SENSOR PARSING OF RUN FAILURE FAILED - {e}<br>{generate_run_url(context)}"
        )

    return body_text


def create_email_alert_subject(context: "RunFailureSensorContext") -> str:
    current_time_str = get_current_time().strftime("%Y-%m-%dT%H:%M:%S")

    schedule_name = context.dagster_run.tags.get(
        "dagster/schedule_name", "UNKNOWN SCHEDULE"
    )
    return f"{current_time_str} Ampere Dagster Run Failed: {schedule_name}"


email_on_run_failure = make_email_on_run_failure_sensor(
    email_from=get_secret("AMPERE_BACKEND_EMAIL_FROM"),
    email_password=get_secret("AMPERE_BACKEND_EMAIL_PW"),
    email_to=get_secret("AMPERE_BACKEND_EMAIL_LIST").split(","),
    monitor_all_code_locations=True,
    default_status=DefaultSensorStatus.RUNNING,
    email_body_fn=create_email_alert_body,
    email_subject_fn=create_email_alert_subject,
)



from dagster import AssetSelection, define_asset_job

github_metrics_daily_4_job = define_asset_job(
    name="github_metrics_daily_4",
    selection=AssetSelection.groups("github_metrics_daily_4"),
)

bigquery_daily_job = define_asset_job(
    name="bigquery_daily", selection=AssetSelection.groups("bigquery_daily")
)



import time
from typing import Any

from dagster import AssetExecutionContext, asset
from dagster_dbt import DbtCliResource, dbt_assets

from ampere.cache_plots import (
    cache_downloads_plots,
    cache_follower_network,
    cache_stargazer_network,
    cache_summary_plots,
    create_follower_network,
    create_stargazer_network,
)
from ampere.common import (
    DeltaTableWriteMode,
    DeltaWriteConfig,
    get_backend_db_con,
    get_model_primary_key,
    write_delta_table,
)
from ampere.get_pypi_downloads import (
    refresh_all_pypi_downloads,
)
from ampere.get_repo_metrics import (
    get_commits,
    get_forks,
    get_issues,
    get_org_user_ids,
    get_pull_requests,
    get_releases,
    get_repos,
    get_stale_followers_user_ids,
    get_stargazers,
    read_repos,
    refresh_followers,
    refresh_github_table,
    refresh_users,
)
from ampere.mirror import copy_backend_to_frontend
from ampere.models import (
    Commit,
    Follower,
    Fork,
    Issue,
    PullRequest,
    Release,
    Repo,
    Stargazer,
    User,
)

from .project import ampere_project

db_path = ampere_project.project_dir.joinpath("data/backend.duckdb")


# DBT
@dbt_assets(manifest=ampere_project.manifest_path)
def ampere_dbt_assets(context: AssetExecutionContext, dbt: DbtCliResource) -> Any:
    yield from dbt.cli(["build"], context=context).stream()


# GITHUB
@asset(
    compute_kind="python",
    key=["repos"],
    group_name="github_metrics_daily_4",
)
def dagster_get_repos(context: AssetExecutionContext) -> None:
    repos = get_repos("mrpowers-io")
    write_delta_table(
        records=repos,
        config=DeltaWriteConfig(
            table_dir="bronze",
            table_name=str(Repo.__tablename__),
            pks=get_model_primary_key(Repo),
            mode=DeltaTableWriteMode.APPEND,
        ),
    )
    context.add_output_metadata({"n_records": len(repos)})


@asset(
    compute_kind="python",
    key=["stargazers"],
    deps=["stg_repos"],
    group_name="github_metrics_daily_4",
)
def dagster_get_stargazers(context: AssetExecutionContext) -> None:
    repos = read_repos(get_backend_db_con())
    owner_name = "mrpowers-io"
    n = refresh_github_table(
        owner_name,
        repos,
        DeltaWriteConfig(
            table_dir="bronze",
            table_name=Stargazer.__tablename__,  # pyright: ignore [reportArgumentType]
            pks=get_model_primary_key(Stargazer),
            mode=DeltaTableWriteMode.APPEND,
        ),
        get_stargazers,
    )
    context.add_output_metadata({"n_records": n})


@asset(
    compute_kind="python",
    key=["forks"],
    deps=["stg_repos"],
    group_name="github_metrics_daily_4",
)
def dagster_get_forks(context: AssetExecutionContext) -> None:
    repos = read_repos(get_backend_db_con())
    owner_name = "mrpowers-io"
    n = refresh_github_table(
        owner_name,
        repos,
        DeltaWriteConfig(
            table_dir="bronze",
            table_name=Fork.__tablename__,  # pyright: ignore [reportArgumentType]
            pks=get_model_primary_key(Fork),
            mode=DeltaTableWriteMode.APPEND,
        ),
        get_forks,
    )
    context.add_output_metadata({"n_records": n})


@asset(
    compute_kind="python",
    key=["releases"],
    deps=["stg_repos"],
    group_name="github_metrics_daily_4",
)
def dagster_get_releases(context: AssetExecutionContext) -> None:
    repos = read_repos(get_backend_db_con())
    owner_name = "mrpowers-io"
    n = refresh_github_table(
        owner_name,
        repos,
        DeltaWriteConfig(
            table_dir="bronze",
            table_name=Release.__tablename__,  # pyright: ignore [reportArgumentType]
            pks=get_model_primary_key(Release),
            mode=DeltaTableWriteMode.APPEND,
        ),
        get_releases,
    )
    context.add_output_metadata({"n_records": n})


@asset(
    compute_kind="python",
    key=["pull_requests"],
    deps=["stg_repos"],
    group_name="github_metrics_daily_4",
)
def dagster_get_pull_requests(context: AssetExecutionContext) -> None:
    repos = read_repos(get_backend_db_con())
    owner_name = "mrpowers-io"
    n = refresh_github_table(
        owner_name,
        repos,
        DeltaWriteConfig(
            table_dir="bronze",
            table_name=PullRequest.__tablename__,  # pyright: ignore [reportArgumentType]
            pks=get_model_primary_key(PullRequest),
            mode=DeltaTableWriteMode.APPEND,
        ),
        get_pull_requests,
    )
    context.add_output_metadata({"n_records": n})


@asset(
    compute_kind="python",
    key=["issues"],
    deps=["stg_repos"],
    group_name="github_metrics_daily_4",
)
def dagster_get_issues(context: AssetExecutionContext) -> None:
    repos = read_repos(get_backend_db_con())
    owner_name = "mrpowers-io"
    n = refresh_github_table(
        owner_name,
        repos,
        DeltaWriteConfig(
            table_dir="bronze",
            table_name=Issue.__tablename__,  # pyright: ignore [reportArgumentType]
            pks=get_model_primary_key(Issue),
            mode=DeltaTableWriteMode.APPEND,
        ),
        get_issues,
    )
    context.add_output_metadata({"n_records": n})


@asset(
    compute_kind="python",
    key=["commits"],
    deps=["stg_repos"],
    group_name="github_metrics_daily_4",
)
def dagster_get_commits(context: AssetExecutionContext) -> None:
    repos = read_repos(get_backend_db_con())
    owner_name = "mrpowers-io"
    n = refresh_github_table(
        owner_name,
        repos,
        DeltaWriteConfig(
            table_dir="bronze",
            table_name=Commit.__tablename__,  # pyright: ignore [reportArgumentType]
            pks=get_model_primary_key(Commit),
            mode=DeltaTableWriteMode.APPEND,
        ),
        get_commits,
    )
    context.add_output_metadata({"n_records": n})


@asset(
    compute_kind="python",
    key=["users"],
    deps=[
        "stg_stargazers",
        "stg_forks",
        "stg_commits",
        "stg_issues",
        "stg_pull_requests",
    ],
    group_name="github_metrics_daily_4",
)
def dagster_get_users(context: AssetExecutionContext) -> None:
    user_ids = get_org_user_ids()
    n = refresh_users(
        user_ids,
        DeltaWriteConfig(
            table_dir="bronze",
            table_name=User.__tablename__,  # pyright: ignore [reportArgumentType]
            pks=get_model_primary_key(User),
            mode=DeltaTableWriteMode.APPEND,
        ),
    )

    context.add_output_metadata({"n_records": n})


@asset(
    compute_kind="python",
    key=["refresh_star_network"],
    deps=["int_network_stargazers", "github_metrics_backend_to_frontend"],
    group_name="github_metrics_daily_4",
)
def dagster_refresh_star_network(context: AssetExecutionContext) -> None:
    start_time = time.time()
    create_stargazer_network()
    cache_stargazer_network()
    context.add_output_metadata({"elapsed time": time.time() - start_time})


@asset(
    compute_kind="python",
    key=["refresh_follower_network"],
    deps=[
        "int_network_follower_details",
        "github_metrics_backend_to_frontend",
        "refresh_star_network",
    ],
    group_name="github_metrics_daily_4",
)
def dagster_refresh_follower_network(context: AssetExecutionContext) -> None:
    start_time = time.time()
    create_follower_network()
    cache_follower_network()
    context.add_output_metadata({"elapsed time": time.time() - start_time})


@asset(
    compute_kind="python",
    key=["followers"],
    deps=["stg_users"],
    group_name="github_metrics_daily_4",
)
def dagster_get_followers(context: AssetExecutionContext) -> None:
    user_ids = get_stale_followers_user_ids("followers")
    n = refresh_followers(
        user_ids,
        DeltaWriteConfig(
            table_dir="bronze",
            table_name=Follower.__tablename__,  # pyright: ignore [reportArgumentType]
            pks=get_model_primary_key(Follower),
            mode=DeltaTableWriteMode.APPEND,
        ),
        "followers",
    )

    context.add_output_metadata({"n_records": n})


@asset(
    compute_kind="python",
    key=["following"],
    deps=["stg_users"],
    group_name="github_metrics_daily_4",
)
def dagster_get_following(context: AssetExecutionContext) -> None:
    user_ids = get_stale_followers_user_ids("following")
    n = refresh_followers(
        user_ids,
        DeltaWriteConfig(
            table_dir="bronze",
            table_name=Follower.__tablename__,  # pyright: ignore [reportArgumentType]
            pks=get_model_primary_key(Follower),
            mode=DeltaTableWriteMode.APPEND,
        ),
        "following",
    )

    context.add_output_metadata({"n_records": n})


@asset(
    compute_kind="python",
    key=["github_metrics_backend_to_frontend"],
    deps=[
        "stg_repos",
        "int_network_stargazers",
        "int_network_follower_details",
        "mart_feed_events",
        "mart_issues",
        "mart_issues_summary",
        "mart_stargazers_pivoted",
        "mart_repo_summary",
    ],
    group_name="github_metrics_daily_4",
)
def github_metrics_table_copy(context: AssetExecutionContext) -> None:
    start_time = time.time()
    copy_backend_to_frontend()
    context.add_output_metadata({"elapsed_time": time.time() - start_time})


@asset(
    compute_kind="python",
    key=["refresh_summary_plots"],
    deps=["github_metrics_backend_to_frontend", "mart_repo_summary"],
    group_name="github_metrics_daily_4",
)
def dagster_refresh_summary_plots(context: AssetExecutionContext) -> None:
    start_time = time.time()
    cache_summary_plots()
    context.add_output_metadata({"elapsed time": time.time() - start_time})


# BIGQUERY
@asset(compute_kind="python", key=["pypi_downloads"], group_name="bigquery_daily")
def dagster_get_pypi_downloads(context: AssetExecutionContext) -> None:
    records_added = refresh_all_pypi_downloads(dry_run=False)
    context.add_output_metadata({"n_records": records_added})


@asset(
    compute_kind="python",
    key=["bigquery_backend_to_frontend"],
    deps=["mart_downloads_summary"],
    group_name="bigquery_daily",
)
def bigquery_table_copy(context: AssetExecutionContext) -> None:
    start_time = time.time()
    copy_backend_to_frontend()
    context.add_output_metadata({"elapsed_time": time.time() - start_time})


@asset(
    compute_kind="python",
    key=["refresh_downloads_plots"],
    deps=["bigquery_backend_to_frontend", "mart_downloads_summary"],
    group_name="bigquery_daily",
)
def dagster_refresh_downloads_plots(context: AssetExecutionContext) -> None:
    start_time = time.time()
    cache_downloads_plots()
    context.add_output_metadata({"elapsed time": time.time() - start_time})


# TESTS
@asset(compute_kind="python", key=["will_pass"], group_name="test")
def dagster_test_run_pass() -> None:
    assert True


@asset(compute_kind="python", key=["will_fail1"], deps=["will_pass"], group_name="test")
def dagster_test_run_fail() -> None:
    raise AssertionError("failed on line 42: this asset always fails. asset1")


@asset(compute_kind="python", key=["will_fail2"], deps=["will_pass"], group_name="test")
def dagster_test_run_fail2() -> None:
    raise AssertionError("failed on line 42: this asset always fails. asset2")



import datetime
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import Callable, Optional

import duckdb
import requests

from ampere.common import (
    DeltaWriteConfig,
    create_header,
    get_backend_db_con,
    get_current_time,
    get_secret,
    timeit,
    write_delta_table,
)
from ampere.models import (
    Commit,
    CommitStats,
    Follower,
    Fork,
    Issue,
    Language,
    PullRequest,
    Release,
    Repo,
    Stargazer,
    User,
)


@dataclass
class APIRequest:
    url: str
    max_requests: int = 50
    max_errors: int = 1
    headers: dict = field(
        default_factory=lambda: {
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {get_secret('GITHUB_TOKEN')}",
            "X-GitHub-Api-Version": "2022-11-28",
        }
    )
    parameters: Optional[dict] = None
    wait_for_quota: bool = True


@dataclass
class APIResponse:
    results: list[dict]
    timestamp: datetime.datetime
    status_code: int


@dataclass
class TaskSleepConfig:
    n_workers: int
    target_adj_pct: float
    request_cpu_time_seconds: float
    task_real_time_seconds: float


def get_rate_limit_reset_sleep_seconds() -> int:
    result = handle_api_response(
        APIRequest(
            url="https://api.github.com/rate_limit",
            max_requests=1,
            max_errors=0,
        )
    )[0].results[0]

    reset = int(result["resources"]["core"]["reset"])
    seconds_to_sleep = reset - int(time.time()) + 1

    print("sleeping for", seconds_to_sleep, "seconds")
    print(
        "requests will resume at",
        datetime.datetime.now() + datetime.timedelta(seconds_to_sleep),
    )
    return seconds_to_sleep


def get_task_sleep_seconds(config: TaskSleepConfig) -> float:
    # https://docs.github.com/en/rest/using-the-rest-api/rate-limits-for-the-rest-api?apiVersion=2022-11-28#about-secondary-rate-limits
    rate_limit_seconds_per_60_seconds = 90
    target_cpu_time_per_60_seconds = (
        rate_limit_seconds_per_60_seconds * config.target_adj_pct
    )
    task_sleep_seconds = (
        1
        / (
            (target_cpu_time_per_60_seconds)
            / (config.request_cpu_time_seconds * 60 * config.n_workers)
        )
        - config.task_real_time_seconds
    ) / config.n_workers
    return task_sleep_seconds


def read_repos(con: duckdb.DuckDBPyConnection) -> list[Repo]:
    repos_dict = con.sql("select * from stg_repos").to_df().to_dict("records")
    repos = [Repo.model_validate(i) for i in repos_dict]
    return repos


def get_latest_commit_timestamp(repo: Repo) -> datetime.datetime | None:
    con = get_backend_db_con()
    query = f"""
        select max(committed_at)
        from stg_commits 
        where repo_id = {repo.repo_id}
        """

    try:
        latest_commit = con.sql(query).fetchone()
    except duckdb.CatalogException as e:
        print(e)
        return None

    if latest_commit is None:
        print(f"{repo.repo_name} has no commits")
        return latest_commit

    return latest_commit[0]


@timeit
def get_org_user_ids() -> list[int]:
    """
    gets unique list of user ids from tracked tables to be added to the `users` table
    if `stg_users` is available, only returns user_ids that have not been refreshed in past `stale_hours` hours
    """
    print("getting user ids from org tables...")
    con = get_backend_db_con()
    stale_hours = 24
    user_ids = (
        con.sql(
            """
            with combined as (
                select user_id from stg_stargazers
                union
                select owner_id as user_id from stg_forks
                union
                select author_id as user_id from stg_commits
                union
                select author_id as user_id from stg_issues
                union
                select author_id as user_id from stg_pull_requests
            )
            select distinct user_id::bigint as user_id
            from combined 
            where user_id is not null
        """
        )
        .to_df()
        .squeeze()
        .tolist()
    )

    try:
        fresh_user_ids = (
            con.sql(
                f"""
                select distinct user_id
                from stg_users
                where retrieved_at >= now() - interval {stale_hours} hour
            """
            )
            .to_df()
            .squeeze()
            .tolist()
        )
        stale_user_ids = sorted(set(user_ids) - set(fresh_user_ids))

    except duckdb.CatalogException as e:
        print(e)
        stale_user_ids = user_ids

    print(f"got {len(stale_user_ids)} stale users")
    return stale_user_ids


@timeit
def get_stale_followers_user_ids(endpoint: str) -> list[int]:
    if endpoint not in ["followers", "following"]:
        raise ValueError("expecting one of ['followers', 'following']")

    con = get_backend_db_con()
    stale_hours = 24
    if endpoint == "followers":
        query = f"""
        select distinct
        a.user_id
        from stg_users a
        left join stg_followers b 
            on a.user_id = b.user_id
        where 
            b.user_id is null
            or b.retrieved_at < now() - interval {stale_hours} hour 
    """

    else:
        query = f"""
        select distinct
        a.user_id
        from stg_users a
        left join stg_followers b 
            on a.user_id = b.follower_id
        where 
            b.follower_id is null
            or b.retrieved_at < now() - interval {stale_hours} hour
        """

    try:
        user_ids = con.sql(query).to_df().squeeze().tolist()
        if isinstance(user_ids, int):
            user_ids = [user_ids]
        print("got stale followers from stg_followers table")
    except Exception as e:
        if "does not exist" in str(e):
            return get_org_user_ids()
        raise Exception(e)

    return user_ids


def handle_api_response(config: APIRequest) -> list[APIResponse]:
    url = config.url
    endpoint = config.url.split("api.github.com")[-1]
    n_requests = 0
    errors = 0

    output: list[APIResponse] = []
    while n_requests < config.max_requests:
        response = requests.get(url=url, headers=config.headers, params=config.parameters)
        n_requests += 1
        print(f"[{endpoint}] requests: {n_requests}")
        # print(f"[{endpoint}] requests: {n_requests}", end="\r", flush=True)

        response_json = response.json()
        if response.status_code == 404:
            return [
                APIResponse(
                    [response_json], get_current_time(), status_code=response.status_code
                )
            ]

        if response.status_code in [403, 429]:
            errors += 1
            if errors > config.max_errors:
                break

            if not config.wait_for_quota:
                # return single response with the rate limit status code to exclude incomplete output downstream
                return [
                    APIResponse(
                        [response_json],
                        get_current_time(),
                        status_code=response.status_code,
                    )
                ]

            response_text = str(response_json)
            if "secondary" in response_text:
                print("hit secondary rate limit")
                time.sleep(60 * errors)
            else:
                time.sleep(get_rate_limit_reset_sleep_seconds())
            continue

        if response.status_code != 200:
            raise ValueError(response.status_code)

        if not isinstance(response_json, list):
            response_json = [response_json]

        output.append(
            APIResponse(response_json, get_current_time(), response.status_code)
        )

        if not response.links:
            break

        # exit if no more continuation tokens
        requests_finished = "next" not in response.links
        if requests_finished:
            break

        url = response.links["next"]["url"]
    return output


def get_forks(owner_name: str, repo: Repo) -> list[Fork]:
    print("getting forks...")
    output = []
    responses = handle_api_response(
        APIRequest(
            url=f"https://api.github.com/repos/{owner_name}/{repo.repo_name}/forks",
            parameters={"per_page": 100},
        )
    )

    for response in responses:
        for result in response.results:
            output.append(
                Fork(
                    repo_id=repo.repo_id,
                    fork_id=result["id"],
                    owner_id=result["owner"]["id"],
                    created_at=result["created_at"],
                    retrieved_at=response.timestamp,
                )
            )
    return output


def get_stargazers(owner_name: str, repo: Repo) -> list[Stargazer]:
    # https://docs.github.com/en/rest/activity/starring?apiVersion=2022-11-28
    print("getting stargazers...")
    output = []
    config = APIRequest(
        url=f"https://api.github.com/repos/{owner_name}/{repo.repo_name}/stargazers",
        parameters={"per_page": 100},
    )
    config.headers["Accept"] = "application/vnd.github.star+json"

    responses = handle_api_response(config)

    for response in responses:
        for result in response.results:
            output.append(
                Stargazer(
                    repo_id=repo.repo_id,
                    user_id=result["user"]["id"],
                    starred_at=result["starred_at"],
                    retrieved_at=response.timestamp,
                )
            )
    return output


def get_repo_language(owner_name: str, repo_name: str) -> list[Language]:
    result = handle_api_response(
        APIRequest(
            url=f"https://api.github.com/repos/{owner_name}/{repo_name}/languages",
            max_requests=1,
            max_errors=0,
        )
    )[0].results[0]

    return [Language(name=k, size_bytes=v) for k, v in result.items()]


def get_repos(org_name: str) -> list[Repo]:
    print("getting repos")
    output = []
    response = handle_api_response(
        APIRequest(
            url=f"https://api.github.com/orgs/{org_name}/repos",
            max_requests=1,
        )
    )[0]

    for result in response.results:
        language = get_repo_language(owner_name=org_name, repo_name=result["name"])
        repo_license = None
        if result["license"] is not None:
            repo_license = result["license"]["name"]
        output.append(
            Repo(
                repo_id=result["id"],
                repo_name=result["name"],
                license=repo_license,
                topics=result["topics"],
                language=language,
                repo_size=result["size"],
                forks_count=result["forks_count"],
                stargazers_count=result["stargazers_count"],
                open_issues_count=result["open_issues_count"],
                pushed_at=result["pushed_at"],
                created_at=result["created_at"],
                updated_at=result["updated_at"],
                retrieved_at=response.timestamp,
            )
        )

    return output


def get_releases(owner_name: str, repo: Repo) -> list[Release]:
    print("getting releases...")
    output = []
    responses = handle_api_response(
        APIRequest(
            url=f"https://api.github.com/repos/{owner_name}/{repo.repo_name}/releases",
            parameters={"per_page": 100},
        )
    )
    for response in responses:
        for result in response.results:
            output.append(
                Release(
                    repo_id=repo.repo_id,
                    release_id=result["id"],
                    release_name=result["name"],
                    tag_name=result["tag_name"],
                    release_body=result["body"],
                    created_at=result["created_at"],
                    published_at=result["published_at"],
                    retrieved_at=response.timestamp,
                )
            )

    return output


def get_commit_stats(
    owner_name: str, repo_name: str, commit_id: str
) -> list[CommitStats]:
    output = []
    result = handle_api_response(
        APIRequest(
            url=f"https://api.github.com/repos/{owner_name}/{repo_name}/commits/{commit_id}",
            max_requests=1,
            max_errors=0,
        )
    )[0].results[0]

    for f in result["files"]:
        output.append(
            CommitStats(
                filename=f["filename"],
                additions=f["additions"],
                deletions=f["deletions"],
                changes=f["changes"],
                status=f["status"],
            )
        )

    return output


def get_commits(owner_name: str, repo: Repo) -> list[Commit]:
    print("getting commits...")
    # by default, sorts by created descending
    output = []
    latest_commit_timestamp = get_latest_commit_timestamp(repo)

    config = APIRequest(
        url=f"https://api.github.com/repos/{owner_name}/{repo.repo_name}/commits",
        parameters={"per_page": 100},
    )

    if latest_commit_timestamp is not None and config.parameters is not None:
        latest_commit_timestamp = latest_commit_timestamp + datetime.timedelta(seconds=1)
        config.parameters["since"] = latest_commit_timestamp

    responses = handle_api_response(config)
    for response in responses:
        for result in response.results:
            commit_stats = get_commit_stats(
                owner_name,
                repo.repo_name,
                result["sha"],
            )

            if result["author"] is not None:
                author_id = result["author"]["id"]
            elif result["committer"] is not None:
                author_id = result["committer"]["id"]
            else:
                author_id = None

            output.append(
                Commit(
                    repo_id=repo.repo_id,
                    commit_id=result["sha"],
                    author_id=author_id,
                    comment_count=result["commit"]["comment_count"],
                    message=result["commit"]["message"],
                    stats=commit_stats,
                    committed_at=result["commit"]["author"]["date"],
                    retrieved_at=response.timestamp,
                )
            )

    return output


def get_pull_requests(owner_name: str, repo: Repo) -> list[PullRequest]:
    print("getting prs...")
    output = []
    responses = handle_api_response(
        APIRequest(
            url=f"https://api.github.com/repos/{owner_name}/{repo.repo_name}/pulls",
            parameters={"per_page": 100, "state": "all"},
        )
    )

    for response in responses:
        for result in response.results:
            output.append(
                PullRequest(
                    repo_id=repo.repo_id,
                    pr_id=result["id"],
                    pr_number=result["number"],
                    pr_title=result["title"],
                    pr_state=result["state"],
                    pr_body=result["body"],
                    author_id=result["user"]["id"],
                    created_at=result["created_at"],
                    updated_at=result["updated_at"],
                    closed_at=result["closed_at"],
                    merged_at=result["merged_at"],
                    retrieved_at=response.timestamp,
                )
            )

    return output


def get_issues(owner_name: str, repo: Repo) -> list[Issue]:
    print("getting issues...")
    output = []
    responses = handle_api_response(
        APIRequest(
            url=f"https://api.github.com/repos/{owner_name}/{repo.repo_name}/issues",
            parameters={"per_page": 100, "state": "all"},
        )
    )
    for response in responses:
        for result in response.results:
            output.append(
                Issue(
                    repo_id=repo.repo_id,
                    issue_id=result["id"],
                    issue_number=result["number"],
                    issue_title=result["title"],
                    issue_body=result["body"],
                    author_id=result["user"]["id"],
                    state=result["state"],
                    state_reason=result["state_reason"],
                    comments_count=result["comments"],
                    created_at=result["created_at"],
                    updated_at=result["updated_at"],
                    closed_at=result["closed_at"],
                    retrieved_at=response.timestamp,
                )
            )
    return output


def get_followers(user_id: int, endpoint: str) -> tuple[list[Follower], bool]:
    # returns tuple of follower list and termination indicator
    if endpoint not in ["followers", "following"]:
        raise ValueError("expecting one of ['followers', 'following']")

    print(f"getting {endpoint} ...")
    output = []
    responses = handle_api_response(
        APIRequest(
            url=f"https://api.github.com/user/{user_id}/{endpoint}",
            max_requests=5000,
            parameters={"per_page": 100},
            wait_for_quota=False,
        )
    )

    skips = 0
    terminate_requests = False
    if endpoint == "followers":
        for response in responses:
            if response.status_code != 200:
                if response.status_code in [403, 429]:
                    terminate_requests = True
                skips += 1
                print(f"skipping {response.status_code} | n skips = {skips}")
                continue
            for result in response.results:
                output.append(
                    Follower(
                        user_id=user_id,
                        follower_id=result["id"],
                        retrieved_at=response.timestamp,
                    )
                )
    else:
        for response in responses:
            if response.status_code != 200:
                if response.status_code in [403, 429]:
                    terminate_requests = True
                skips += 1
                print(f"skipping {response.status_code} | n skips = {skips}")
                continue
            for result in response.results:
                output.append(
                    Follower(
                        user_id=result["id"],
                        follower_id=user_id,
                        retrieved_at=response.timestamp,
                    )
                )

    return output, terminate_requests


def get_user(user_id: int) -> Optional[User]:
    response = handle_api_response(
        APIRequest(
            url=f"https://api.github.com/user/{user_id}",
            max_requests=3,
        )
    )[0]

    time.sleep(
        get_task_sleep_seconds(
            TaskSleepConfig(
                n_workers=2,
                target_adj_pct=0.9,
                request_cpu_time_seconds=0.25,
                task_real_time_seconds=0.3,
            )
        )
    )
    if response.status_code != 200:
        return None

    results = response.results[0]
    return User(
        user_id=results["id"],
        user_name=results["login"],
        full_name=results["name"],
        company=results["company"],
        avatar_url=results["avatar_url"],
        repos_count=results["public_repos"],
        followers_count=results["followers"],
        following_count=results["following"],
        created_at=results["created_at"],
        updated_at=results["updated_at"],
        retrieved_at=response.timestamp,
    )


def refresh_github_table(
    owner_name: str,
    repos: list[Repo],
    config: DeltaWriteConfig,
    get_func: Callable,
) -> int:
    all_results = []
    for i, repo in enumerate(repos, 1):
        header_text = f"[{i:02d}/{len(repos):02d}] {repo.repo_name}"
        print(create_header(80, header_text, True, "-"))

        results = get_func(owner_name, repo)
        print(f"obtained {len(results)} records")
        all_results.extend(results)

    if len(all_results) == 0:
        print("no records to write")
        return 0

    write_delta_table(records=all_results, config=config)
    return len(all_results)


def refresh_users(
    user_ids: list[int],
    config: DeltaWriteConfig,
) -> int:
    if len(user_ids) == 0:
        print("no user ids to process. exiting early")
        return 0

    if any(not isinstance(i, int) for i in user_ids):
        raise TypeError("expecting user id of type `int`")

    start_time = time.time()
    with ThreadPoolExecutor(max_workers=2) as executor:
        raw_results = list(executor.map(get_user, user_ids))

    all_results = [i for i in raw_results if i is not None]
    if len(all_results) == 0:
        print("no results obtained")
        return 0

    elapsed_time = time.time() - start_time
    avg_time_per_user = elapsed_time / len(user_ids)

    print(f"elapsed time: {elapsed_time:.2f} seconds")
    print(f"average time per user: {avg_time_per_user:.2f} seconds")

    write_delta_table(records=all_results, config=config)
    return len(all_results)


def refresh_followers(
    user_ids: list[int], config: DeltaWriteConfig, endpoint: str
) -> int:
    all_results = []
    start_time = time.time()
    if len(user_ids) == 0:
        print("no user ids to process. exiting early")
        return 0

    if any(not isinstance(i, int) for i in user_ids):
        raise TypeError("expecting user id of type `int`")

    for i, user_id in enumerate(user_ids, 1):
        header_text = f"[{i:04d}/{len(user_ids):04d}] {user_id}"
        print(create_header(80, header_text, True, "-"))
        result, terminate_requests = get_followers(user_id, endpoint)

        if terminate_requests:
            break

        all_results.extend(result)

    elapsed_time = time.time() - start_time
    valid_results = [i for i in all_results if i is not None]
    if len(valid_results) == 0:
        print("no results obtained")
        return 0

    avg_time_per_user = elapsed_time / len(user_ids)

    print(f"elapsed time: {elapsed_time:.2f} seconds")
    print(f"average time per user: {avg_time_per_user:.2f} seconds")

    write_delta_table(records=valid_results, config=config)
    return len(valid_results)



import datetime
from typing import Annotated, Optional

import typer

from ampere.common import get_backend_db_con
from ampere.get_pypi_downloads import add_backfill_to_table

app = typer.Typer()


@app.command()
def backfill(
    repo: Annotated[str, typer.Option()],
    min_date: Optional[str] = None,
    max_date: Optional[str] = None,
    repo_dependency: Optional[str] = None,
    max_days_per_chunk: int = 15,
    dry_run: bool = True,
) -> None:
    if min_date is None and repo_dependency is None:
        raise ValueError(
            "expecting either `min_date` or `repo_dependency` to be provided"
        )

    if min_date is not None:
        min_date_dt = datetime.datetime.strptime(min_date, "%Y-%m-%d").replace(
            tzinfo=datetime.timezone.utc
        )
    else:
        con = get_backend_db_con()
        min_date_dt = con.sql(
            f"select created_at from stg_repos where repo_name = '{repo_dependency}'"
        ).fetchall()[0][0]

    if max_date is not None:
        max_date_dt = datetime.datetime.strptime(max_date, "%Y-%m-%d").replace(
            tzinfo=datetime.timezone.utc
        )
    else:
        max_date_dt = None

    add_backfill_to_table(
        repo,
        min_date_dt,
        max_date_dt,
        max_days_per_chunk,
        dry_run,
    )


if __name__ == "__main__":
    app()



from pathlib import Path

import typer
from deltalake import DeltaTable

app = typer.Typer()


@app.command()
def delete(table_name: str, predicate: str, dry_run: bool = True) -> None:
    tbl_path = Path(__file__).parents[1] / "data" / "bronze" / table_name
    delta_table = DeltaTable(tbl_path)

    print(predicate)
    if dry_run:
        return
    delta_table.delete(predicate)


if __name__ == "__main__":
    app()
